@@ -1,6 +1,6 @@
 /* BSD Zero Clause License */
 
-/* Copyright (C) 2022-2024 mintsuki and contributors.
+/* Copyright (C) 2022-2025 Mintsuki and contributors.
  *
  * Permission to use, copy, modify, and/or distribute this software for any
  * purpose with or without fee is hereby granted.
@@ -31,6 +31,14 @@ extern "C" {
 #  define LIMINE_PTR(TYPE) TYPE
 #endif
 
+#ifndef LIMINE_API_REVISION
+#  define LIMINE_API_REVISION 0
+#endif
+
+#if LIMINE_API_REVISION > 3
+#  error "limine.h API revision unsupported"
+#endif
+
 #ifdef __GNUC__
 #  define LIMINE_DEPRECATED __attribute__((__deprecated__))
 #  define LIMINE_DEPRECATED_IGNORE_START \
@@ -78,7 +86,11 @@ struct limine_file {
     LIMINE_PTR(void *) address;
     uint64_t size;
     LIMINE_PTR(char *) path;
+#if LIMINE_API_REVISION >= 3
+    LIMINE_PTR(char *) string;
+#else
     LIMINE_PTR(char *) cmdline;
+#endif
     uint32_t media_type;
     uint32_t unused;
     uint32_t tftp_ip;
@@ -106,6 +118,21 @@ struct limine_bootloader_info_request {
     LIMINE_PTR(struct limine_bootloader_info_response *) response;
 };
 
+/* Executable command line */
+
+#define LIMINE_EXECUTABLE_CMDLINE_REQUEST { LIMINE_COMMON_MAGIC, 0x4b161536e598651e, 0xb390ad4a2f1f303a }
+
+struct limine_executable_cmdline_response {
+    uint64_t revision;
+    LIMINE_PTR(char *) cmdline;
+};
+
+struct limine_executable_cmdline_request {
+    uint64_t id[4];
+    uint64_t revision;
+    LIMINE_PTR(struct limine_executable_cmdline_response *) response;
+};
+
 /* Firmware type */
 
 #define LIMINE_FIRMWARE_TYPE_REQUEST { LIMINE_COMMON_MAGIC, 0x8c2f75d90bef28a8, 0x7045a4688eac00c3 }
@@ -113,6 +140,7 @@ struct limine_bootloader_info_request {
 #define LIMINE_FIRMWARE_TYPE_X86BIOS 0
 #define LIMINE_FIRMWARE_TYPE_UEFI32 1
 #define LIMINE_FIRMWARE_TYPE_UEFI64 2
+#define LIMINE_FIRMWARE_TYPE_SBI 3
 
 struct limine_firmware_type_response {
     uint64_t revision;
@@ -328,19 +356,29 @@ struct LIMINE_DEPRECATED limine_5_level_paging_request {
 
 LIMINE_DEPRECATED_IGNORE_END
 
-/* SMP */
+/* MP */
 
-#define LIMINE_SMP_REQUEST { LIMINE_COMMON_MAGIC, 0x95a67b819a1b857e, 0xa0b61b723b6a73e0 }
+#if LIMINE_API_REVISION >= 1
+#  define LIMINE_MP_REQUEST { LIMINE_COMMON_MAGIC, 0x95a67b819a1b857e, 0xa0b61b723b6a73e0 }
+#  define LIMINE_MP(TEXT) limine_mp_##TEXT
+#else
+#  define LIMINE_SMP_REQUEST { LIMINE_COMMON_MAGIC, 0x95a67b819a1b857e, 0xa0b61b723b6a73e0 }
+#  define LIMINE_MP(TEXT) limine_smp_##TEXT
+#endif
 
-struct limine_smp_info;
+struct LIMINE_MP(info);
 
-typedef void (*limine_goto_address)(struct limine_smp_info *);
+typedef void (*limine_goto_address)(struct LIMINE_MP(info) *);
 
 #if defined (__x86_64__) || defined (__i386__)
 
-#define LIMINE_SMP_X2APIC (1 << 0)
+#if LIMINE_API_REVISION >= 1
+#  define LIMINE_MP_X2APIC (1 << 0)
+#else
+#  define LIMINE_SMP_X2APIC (1 << 0)
+#endif
 
-struct limine_smp_info {
+struct LIMINE_MP(info) {
     uint32_t processor_id;
     uint32_t lapic_id;
     uint64_t reserved;
@@ -348,17 +386,17 @@ struct limine_smp_info {
     uint64_t extra_argument;
 };
 
-struct limine_smp_response {
+struct LIMINE_MP(response) {
     uint64_t revision;
     uint32_t flags;
     uint32_t bsp_lapic_id;
     uint64_t cpu_count;
-    LIMINE_PTR(struct limine_smp_info **) cpus;
+    LIMINE_PTR(struct LIMINE_MP(info) **) cpus;
 };
 
 #elif defined (__aarch64__)
 
-struct limine_smp_info {
+struct LIMINE_MP(info) {
     uint32_t processor_id;
     uint32_t reserved1;
     uint64_t mpidr;
@@ -367,17 +405,17 @@ struct limine_smp_info {
     uint64_t extra_argument;
 };
 
-struct limine_smp_response {
+struct LIMINE_MP(response) {
     uint64_t revision;
     uint64_t flags;
     uint64_t bsp_mpidr;
     uint64_t cpu_count;
-    LIMINE_PTR(struct limine_smp_info **) cpus;
+    LIMINE_PTR(struct LIMINE_MP(info) **) cpus;
 };
 
 #elif defined (__riscv) && (__riscv_xlen == 64)
 
-struct limine_smp_info {
+struct LIMINE_MP(info) {
     uint64_t processor_id;
     uint64_t hartid;
     uint64_t reserved;
@@ -385,33 +423,33 @@ struct limine_smp_info {
     uint64_t extra_argument;
 };
 
-struct limine_smp_response {
+struct LIMINE_MP(response) {
     uint64_t revision;
     uint64_t flags;
     uint64_t bsp_hartid;
     uint64_t cpu_count;
-    LIMINE_PTR(struct limine_smp_info **) cpus;
+    LIMINE_PTR(struct LIMINE_MP(info) **) cpus;
 };
 
 #elif defined (__loongarch__) && (__loongarch_grlen == 64)
 
-struct limine_smp_info {
+struct LIMINE_MP(info) {
     uint64_t reserved;
 };
 
-struct limine_smp_response {
+struct LIMINE_MP(response) {
     uint64_t cpu_count;
-    LIMINE_PTR(struct limine_smp_info **) cpus;
+    LIMINE_PTR(struct LIMINE_MP(info) **) cpus;
 };
 
 #else
 #error Unknown architecture
 #endif
 
-struct limine_smp_request {
+struct LIMINE_MP(request) {
     uint64_t id[4];
     uint64_t revision;
-    LIMINE_PTR(struct limine_smp_response *) response;
+    LIMINE_PTR(struct LIMINE_MP(response) *) response;
     uint64_t flags;
 };
 
@@ -425,7 +463,11 @@ struct limine_smp_request {
 #define LIMINE_MEMMAP_ACPI_NVS               3
 #define LIMINE_MEMMAP_BAD_MEMORY             4
 #define LIMINE_MEMMAP_BOOTLOADER_RECLAIMABLE 5
-#define LIMINE_MEMMAP_KERNEL_AND_MODULES     6
+#if LIMINE_API_REVISION >= 2
+#  define LIMINE_MEMMAP_EXECUTABLE_AND_MODULES 6
+#else
+#  define LIMINE_MEMMAP_KERNEL_AND_MODULES 6
+#endif
 #define LIMINE_MEMMAP_FRAMEBUFFER            7
 
 struct limine_memmap_entry {
@@ -463,19 +505,39 @@ struct limine_entry_point_request {
     LIMINE_PTR(limine_entry_point) entry;
 };
 
-/* Kernel File */
+/* Executable File */
 
-#define LIMINE_KERNEL_FILE_REQUEST { LIMINE_COMMON_MAGIC, 0xad97e90e83f1ed67, 0x31eb5d1c5ff23b69 }
+#if LIMINE_API_REVISION >= 2
+#  define LIMINE_EXECUTABLE_FILE_REQUEST { LIMINE_COMMON_MAGIC, 0xad97e90e83f1ed67, 0x31eb5d1c5ff23b69 }
+#else
+#  define LIMINE_KERNEL_FILE_REQUEST { LIMINE_COMMON_MAGIC, 0xad97e90e83f1ed67, 0x31eb5d1c5ff23b69 }
+#endif
 
+#if LIMINE_API_REVISION >= 2
+struct limine_executable_file_response {
+#else
 struct limine_kernel_file_response {
+#endif
     uint64_t revision;
+#if LIMINE_API_REVISION >= 2
+    LIMINE_PTR(struct limine_file *) executable_file;
+#else
     LIMINE_PTR(struct limine_file *) kernel_file;
+#endif
 };
 
+#if LIMINE_API_REVISION >= 2
+struct limine_executable_file_request {
+#else
 struct limine_kernel_file_request {
+#endif
     uint64_t id[4];
     uint64_t revision;
+#if LIMINE_API_REVISION >= 2
+    LIMINE_PTR(struct limine_executable_file_response *) response;
+#else
     LIMINE_PTR(struct limine_kernel_file_response *) response;
+#endif
 };
 
 /* Module */
@@ -487,7 +549,11 @@ struct limine_kernel_file_request {
 
 struct limine_internal_module {
     LIMINE_PTR(const char *) path;
+#if LIMINE_API_REVISION >= 3
+    LIMINE_PTR(const char *) string;
+#else
     LIMINE_PTR(const char *) cmdline;
+#endif
     uint64_t flags;
 };
 
@@ -513,7 +579,11 @@ struct limine_module_request {
 
 struct limine_rsdp_response {
     uint64_t revision;
+#if LIMINE_API_REVISION >= 1
+    uint64_t address;
+#else
     LIMINE_PTR(void *) address;
+#endif
 };
 
 struct limine_rsdp_request {
@@ -528,8 +598,13 @@ struct limine_rsdp_request {
 
 struct limine_smbios_response {
     uint64_t revision;
+#if LIMINE_API_REVISION >= 1
+    uint64_t entry_32;
+    uint64_t entry_64;
+#else
     LIMINE_PTR(void *) entry_32;
     LIMINE_PTR(void *) entry_64;
+#endif
 };
 
 struct limine_smbios_request {
@@ -544,7 +619,11 @@ struct limine_smbios_request {
 
 struct limine_efi_system_table_response {
     uint64_t revision;
+#if LIMINE_API_REVISION >= 1
+    uint64_t address;
+#else
     LIMINE_PTR(void *) address;
+#endif
 };
 
 struct limine_efi_system_table_request {
@@ -571,35 +650,71 @@ struct limine_efi_memmap_request {
     LIMINE_PTR(struct limine_efi_memmap_response *) response;
 };
 
-/* Boot time */
+/* Date at boot */
 
-#define LIMINE_BOOT_TIME_REQUEST { LIMINE_COMMON_MAGIC, 0x502746e184c088aa, 0xfbc5ec83e6327893 }
+#if LIMINE_API_REVISION >= 3
+#  define LIMINE_DATE_AT_BOOT_REQUEST { LIMINE_COMMON_MAGIC, 0x502746e184c088aa, 0xfbc5ec83e6327893 }
+#else
+#  define LIMINE_BOOT_TIME_REQUEST { LIMINE_COMMON_MAGIC, 0x502746e184c088aa, 0xfbc5ec83e6327893 }
+#endif
 
+#if LIMINE_API_REVISION >= 3
+struct limine_date_at_boot_response {
+#else
 struct limine_boot_time_response {
+#endif
     uint64_t revision;
+#if LIMINE_API_REVISION >= 3
+    int64_t timestamp;
+#else
     int64_t boot_time;
+#endif
 };
 
+#if LIMINE_API_REVISION >= 3
+struct limine_date_at_boot_request {
+#else
 struct limine_boot_time_request {
+#endif
     uint64_t id[4];
     uint64_t revision;
+#if LIMINE_API_REVISION >= 3
+    LIMINE_PTR(struct limine_date_at_boot_response *) response;
+#else
     LIMINE_PTR(struct limine_boot_time_response *) response;
+#endif
 };
 
-/* Kernel address */
+/* Executable address */
 
-#define LIMINE_KERNEL_ADDRESS_REQUEST { LIMINE_COMMON_MAGIC, 0x71ba76863cc55f63, 0xb2644a48c516a487 }
+#if LIMINE_API_REVISION >= 2
+#  define LIMINE_EXECUTABLE_ADDRESS_REQUEST { LIMINE_COMMON_MAGIC, 0x71ba76863cc55f63, 0xb2644a48c516a487 }
+#else
+#  define LIMINE_KERNEL_ADDRESS_REQUEST { LIMINE_COMMON_MAGIC, 0x71ba76863cc55f63, 0xb2644a48c516a487 }
+#endif
 
+#if LIMINE_API_REVISION >= 2
+struct limine_executable_address_response {
+#else
 struct limine_kernel_address_response {
+#endif
     uint64_t revision;
     uint64_t physical_base;
     uint64_t virtual_base;
 };
 
+#if LIMINE_API_REVISION >= 2
+struct limine_executable_address_request {
+#else
 struct limine_kernel_address_request {
+#endif
     uint64_t id[4];
     uint64_t revision;
+#if LIMINE_API_REVISION >= 2
+    LIMINE_PTR(struct limine_executable_address_response *) response;
+#else
     LIMINE_PTR(struct limine_kernel_address_response *) response;
+#endif
 };
 
 /* Device Tree Blob */
@@ -617,6 +732,38 @@ struct limine_dtb_request {
     LIMINE_PTR(struct limine_dtb_response *) response;
 };
 
+/* RISC-V Boot Hart ID */
+
+#define LIMINE_RISCV_BSP_HARTID_REQUEST { LIMINE_COMMON_MAGIC, 0x1369359f025525f9, 0x2ff2a56178391bb6 }
+
+struct limine_riscv_bsp_hartid_response {
+    uint64_t revision;
+    uint64_t bsp_hartid;
+};
+
+struct limine_riscv_bsp_hartid_request {
+    uint64_t id[4];
+    uint64_t revision;
+    LIMINE_PTR(struct limine_riscv_bsp_hartid_response *) response;
+};
+
+/* Bootloader Performance */
+
+#define LIMINE_BOOTLOADER_PERFORMANCE_REQUEST { LIMINE_COMMON_MAGIC, 0x6b50ad9bf36d13ad, 0xdc4c7e88fc759e17 }
+
+struct limine_bootloader_performance_response {
+    uint64_t revision;
+    uint64_t reset_usec;
+    uint64_t init_usec;
+    uint64_t exec_usec;
+};
+
+struct limine_bootloader_performance_request {
+    uint64_t id[4];
+    uint64_t revision;
+    LIMINE_PTR(struct limine_bootloader_performance_response *) response;
+};
+
 #ifdef __cplusplus
 }
 #endif
